# Janos version 1.2.1
# JanosParser.py
# by Antoine Rauzy
# Copyrights (c) 2018-2022 AltaRica Association
#
# This file is part of Janos, a pedagogical stochastic simulator.
# Janos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Janos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Janos.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Models
#   3.1) Elements
#   3.2) Variables and equations

# 1) Imported modules
# -------------------

from S2MLCore import *

from JanosModels import *
from JanosModelLexer import *

# 2) Creator
# ----------

class JanosModelParser(S2MLModelParser):
  def __init__(self, engine):
    S2MLModelParser.__init__(self, engine)

# 3) Models
# ---------

# 3.1) Elements
# -------------

  def _ParseBlockElement(self, container):
    element = None
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token==None:
      return element
    elif token.GetType()==JanosToken.KEYWORD_ASSERTION:
      element = self._ParseEquationDeclaration(container)
    elif token.GetType()==JanosToken.KEYWORD_PARAMETER:
      element = self._ParseParameter(container)
    elif token.GetType()==JanosToken.KEYWORD_OBSERVER:
      element = self._ParseObserver(container)
    elif token.GetType()==JanosToken.KEYWORD_BLOCK:
      element = self._ParseBlock(container)
    elif token.GetType()==JanosToken.KEYWORD_EXTENDS:
      element = self._ParseExtendsDirective(container)
    elif token.GetType()==JanosToken.KEYWORD_EMBEDS:
      element = self._ParseEmbedsDirective(container)
    elif token.GetType()==JanosToken.KEYWORD_CLONES:
      element = self._ParseClonesDirective(container)
    elif token.GetType()==JanosToken.KEYWORD_BOOLEAN:
      element = self._ParseVariableDeclaration(container)
    elif token.GetType()==JanosToken.KEYWORD_INTEGER:
      element = self._ParseVariableDeclaration(container)
    elif token.GetType()==JanosToken.KEYWORD_REAL:
      element = self._ParseVariableDeclaration(container)
    elif token.GetType()==JanosToken.KEYWORD_SYMBOL:
      element = self._ParseVariableDeclaration(container)
    elif token.GetType()==JanosToken.TOKEN_LBRACE:
      element = self._ParseVariableDeclaration(container)
    elif token.GetType()==JanosToken.TOKEN_LBRACK:
      element = self._ParseVariableDeclaration(container)
    elif token.GetType()==JanosToken.TOKEN_LPAR:
      element = self._ParseFunctionCall(container)
    else:
      instantiatedElementIdentifier = self._ParseIdentifier()
      if instantiatedElementIdentifier==None:
        return None
      instantiatedElement = self._model.LookForIdentifiedElement(instantiatedElementIdentifier)
      if instantiatedElement==None:
        self._RaiseParseError(token, "Undeclared domain or class %s" % instantiatedElementIdentifier)
        return None
      self._currentTokenIndex -= 1
      if JanosModel.IsDomainDeclaration(instantiatedElement):
        element = self._ParseVariableDeclaration(container)
      else:
        element = self._ParseClassInstance(container)
    return element


# 3.2) Variables and equations
# ----------------------------

  def _ParseVariableDeclaration(self, container):
    variables = []
    domain = self._ParseDomain()
    if domain==None:
      return None
    while True:
      path = self._ParsePath()
      if path==None:
        return None
      variable = self._model.NewVariable(domain, path, container)
      variables.append(variable)
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.GetType()==JanosToken.TOKEN_COMMA:
        self._currentTokenIndex += 1
      else:
        break
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()==JanosToken.TOKEN_LPAR:
      variables = self._ParseAttributes(variables)
      if variables==None:
        return None
    token = self._ParseGivenToken(JanosToken.TOKEN_SEMICOLON)
    if token==None:
      return None
    return variables

  def _ParseEquationDeclaration(self, container):
    token = self._ParseGivenToken(JanosToken.KEYWORD_ASSERTION)
    if token==None:
      return None
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token==None or (token.GetType()!=JanosToken.TOKEN_IDENTIFIER and token.GetType()!=JanosToken.TOKEN_PATH):
        break
      equation = self._ParseEquation(container)
      if equation==None:
        # self._RecoverError(JanosToken.KEYWORD_END)
        return None
    return container

  def _ParseEquation(self, container):
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()==JanosToken.TOKEN_LPAR:
      return self._ParseConnection(container)
    leftMemberPath = self._ParsePath()
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()==JanosToken.TOKEN_ASSIGN:
      self._currentTokenIndex += 1
      rightMember = self._ParseStochasticExpression()
      if rightMember==None:
        return None
      token = self._ParseGivenToken(JanosToken.TOKEN_SEMICOLON)
      if token==None:
        return None
      leftMember = S2MLStochasticExpression.NewReference(leftMemberPath)
      return self._model.NewAssignment(leftMember, rightMember, container)
    token = self._ParseGivenToken(JanosToken.TOKEN_DOUBLE_ASSIGN)
    if token==None:
      return None
    rightMemberPath = self._ParsePath()
    if rightMemberPath==None:
      return None
    token = self._ParseGivenToken(JanosToken.TOKEN_SEMICOLON)
    if token==None:
      return None
    leftMember = S2MLStochasticExpression.NewReference(leftMemberPath)
    rightMember = S2MLStochasticExpression.NewReference(rightMemberPath)
    return self._model.NewDoubleAssignment(leftMember, rightMember, container)

  def _ParseObserver(self, container):
    if not self._model.CanComposeObserver(container):
      self._RaiseParseError(token, "Observer declarations are not authorized here")
      return None
    token = self._ParseGivenToken(JanosToken.KEYWORD_OBSERVER)
    if token==None:
      return None
    domain = self._ParseDomain()
    if domain==None:
      return None
    path = self._ParsePath()
    if path==None:
      return None
    token = self._ParseGivenToken(JanosToken.TOKEN_EQUAL)
    if token==None:
      return None
    expression = self._ParseStochasticExpression()
    if expression==None:
      return None
    token = self._ParseGivenToken(JanosToken.TOKEN_SEMICOLON)
    if token==None:
      return None
    observer = self._model.NewObserver(domain, path, container)
    observer.SetDefinition(expression)
    return observer
