# S2MLCore version 1.1.1
# S2MLModelLexer.py
# by Antoine Rauzy
# Copyrights (c) 2018-2022 AltaRica Association
#
# This file is part of S2MLCore.
# S2MLCore is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# S2MLCore is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with S2MLCore.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Management of actions
# 4) Management of keywords
# 5) Main loop
# 6) Auxiliary functions


# 1) Imported modules
# -------------------

import string
import sys
import re

from S2MLToken import *

# 2) Creator
# ----------

class S2MLModelLexer:
  STATE_END_OF_FILE = 0
  STATE_INITIAL = 1
  STATE_IDENTIFIER = 10
  STATE_PATH_1 = 11
  STATE_PATH_2 = 12
  STATE_NUMBER_1 = 20
  STATE_NUMBER_2 = 21
  STATE_NUMBER_3 = 22
  STATE_NUMBER_4 = 23
  STATE_NUMBER_5 = 24
  STATE_NUMBER_6 = 25
  STATE_NUMBER_7 = 26
  STATE_NUMBER_8 = 27
  STATE_SINGLE_LINE_COMMENT = 40
  STATE_MULTILINE_COMMENT_1 = 50
  STATE_MULTILINE_COMMENT_2 = 51

  def __init__(self, engine):
    from S2MLEngine import S2MLEngine
    self._engine = engine
    self._actions = dict()
    self.RegisterActions()
    self._codesOfKeywords = dict()
    self._stringsOfCodes = dict()
    self.RegisterTokens()
    self.RegisterKeywords()
    self.keepComments = False
    self.fileName = None
    self.file = None
    self.lineNumber = 0
    self.columnNumber = 0
    self.line = None
    self.endOfFile = 0
    self.tokenLineNumber = 0
    self.tokenColumnNumber = 0
    self.tokenString = None
    self.readChars = []
    self.tokens = []
    self.state = -1

# 3) Management of actions
# ------------------------

  def RegisterActions(self):
    self.RegisterAction("(", S2MLModelLexer.ActionLPAR)
    self.RegisterAction(")", S2MLModelLexer.ActionRPAR)
    self.RegisterAction("{", S2MLModelLexer.ActionLBRACE)
    self.RegisterAction("}", S2MLModelLexer.ActionRBRACE)
    self.RegisterAction("[", S2MLModelLexer.ActionLBRACK)
    self.RegisterAction("]", S2MLModelLexer.ActionRBRACK)
    self.RegisterAction(";", S2MLModelLexer.ActionSemicolon)
    self.RegisterAction(":", S2MLModelLexer.ActionColon)
    self.RegisterAction(",", S2MLModelLexer.ActionComma)
    self.RegisterAction("=", S2MLModelLexer.ActionEqual)
    self.RegisterAction("+", S2MLModelLexer.ActionPlus)
    self.RegisterAction("-", S2MLModelLexer.ActionMinus)
    self.RegisterAction("*", S2MLModelLexer.ActionStar)
    self.RegisterAction("/", S2MLModelLexer.ActionSlash)
    self.RegisterAction("<", S2MLModelLexer.ActionLess)
    self.RegisterAction(">", S2MLModelLexer.ActionGreater)
    self.RegisterAction("!", S2MLModelLexer.ActionExclamationMark)

  def RegisterAction(self, c, action):
    self._actions[c] = action

  def GetAction(self, c):
    return self._actions.get(c, None)

  def ActionLPAR(self, c):
    self.tokenString = c
    self.NewToken(S2MLToken.TOKEN_LPAR)

  def ActionRPAR(self, c):
    self.tokenString = c
    self.NewToken(S2MLToken.TOKEN_RPAR)

  def ActionLBRACE(self, c):
    self.tokenString = c
    self.NewToken(S2MLToken.TOKEN_LBRACE)

  def ActionRBRACE(self, c):
    self.tokenString = c
    self.NewToken(S2MLToken.TOKEN_RBRACE)

  def ActionLBRACK(self, c):
    self.tokenString = c
    self.NewToken(S2MLToken.TOKEN_LBRACK)

  def ActionRBRACK(self, c):
    self.tokenString = c
    self.NewToken(S2MLToken.TOKEN_RBRACK)

  def ActionSemicolon(self, c):
    self.tokenString = c
    self.NewToken(S2MLToken.TOKEN_SEMICOLON)

  def ActionColon(self, c):
    self.tokenString = c
    self.NewToken(S2MLToken.TOKEN_COLON)

  def ActionComma(self, c):
    self.tokenString = c
    self.NewToken(S2MLToken.TOKEN_COMMA)

  def ActionEqual(self, c):
    c = self.GetChar()
    if c=="=":
      self.tokenString = "=="
      self.NewToken(S2MLToken.TOKEN_EQ)
    else:
      self.PutCharBack(c)
      self.tokenString = "="
      self.NewToken(S2MLToken.TOKEN_EQUAL)

  def ActionStar(self, c):
    self.tokenString = c
    self.NewToken(S2MLToken.TOKEN_STAR)

  def ActionLess(self, c):
    c = self.GetChar()
    if c=="=":
      self.tokenString = "<="
      self.NewToken(S2MLToken.TOKEN_LEQ)
    else:
      self.PutCharBack(c)
      self.tokenString = "<"
      self.NewToken(S2MLToken.TOKEN_LT)

  def ActionGreater(self, c):
    c = self.GetChar()
    if c=="=":
      self.tokenString = ">="
      self.NewToken(S2MLToken.TOKEN_GEQ)
    else:
      self.PutCharBack(c)
      self.tokenString = ">"
      self.NewToken(S2MLToken.TOKEN_GT)

  def ActionExclamationMark(self, c):
    self.tokenString = c
    if c=="=":
      self.tokenString = "!="
      self.NewToken(S2MLToken.TOKEN_DF)
    else:
      self.PutCharBack(c)
      self.tokenString = "!"
      self.NewToken(S2MLToken.TOKEN_EXCLAMATION_MARK)

  def ActionPlus(self, c):
    c = self.GetChar()
    if c.isdigit():
      self.tokenString = "+" + c
      self.state = S2MLModelLexer.STATE_NUMBER_1
    else:
      self.tokenString = "+"
      self.PutCharBack(c)
      self.NewToken(S2MLToken.TOKEN_PLUS)

  def ActionMinus(self, c):
    c = self.GetChar()
    if c.isdigit():
      self.tokenString = "-" + c
      self.state = S2MLModelLexer.STATE_NUMBER_2
    else:
      self.tokenString = "-"
      self.PutCharBack(c)
      self.NewToken(S2MLToken.TOKEN_MINUS)

  def ActionSlash(self, c):
    c = self.GetChar()
    if c=="/":
      self.tokenString = "//"
      self.state = S2MLModelLexer.STATE_SINGLE_LINE_COMMENT
    elif c=="*":
      self.tokenString = "/*"
      self.state = S2MLModelLexer.STATE_MULTILINE_COMMENT_1
    else:
      self.tokenString = "/"
      self.PutCharBack(c)
      self.NewToken(S2MLToken.TOKEN_SLASH)

# 4) Management of keywords
# -------------------------

  def RegisterTokens(self):
    self.RegisterToken("(", S2MLToken.TOKEN_LPAR)
    self.RegisterToken(")", S2MLToken.TOKEN_RPAR)
    self.RegisterToken("{", S2MLToken.TOKEN_LBRACE)
    self.RegisterToken("}", S2MLToken.TOKEN_RBRACE)
    self.RegisterToken(";", S2MLToken.TOKEN_SEMICOLON)
    self.RegisterToken(":", S2MLToken.TOKEN_COLON)
    self.RegisterToken(",", S2MLToken.TOKEN_COMMA)
    self.RegisterToken("=", S2MLToken.TOKEN_EQUAL)
    self.RegisterToken("Integer", S2MLToken.TOKEN_INTEGER)
    self.RegisterToken("Real", S2MLToken.TOKEN_REAL)
    self.RegisterToken("identifier", S2MLToken.TOKEN_IDENTIFIER)
    self.RegisterToken("path", S2MLToken.TOKEN_PATH)
    self.RegisterToken("comment", S2MLToken.TOKEN_COMMENT)

  def RegisterToken(self, string, code):
    self._stringsOfCodes[code] = string

  def RegisterKeywords(self):
    # S2ML
    self.RegisterKeyword("domain", S2MLToken.KEYWORD_DOMAIN)
    self.RegisterKeyword("block", S2MLToken.KEYWORD_BLOCK)
    self.RegisterKeyword("class", S2MLToken.KEYWORD_CLASS)
    self.RegisterKeyword("end", S2MLToken.KEYWORD_END)
    self.RegisterKeyword("clones", S2MLToken.KEYWORD_CLONES)
    self.RegisterKeyword("as", S2MLToken.KEYWORD_AS)
    self.RegisterKeyword("embeds", S2MLToken.KEYWORD_EMBEDS)
    self.RegisterKeyword("extends", S2MLToken.KEYWORD_EXTENDS)
    self.RegisterKeyword("function", S2MLToken.KEYWORD_FUNCTION)
    self.RegisterKeyword("operator", S2MLToken.KEYWORD_OPERATOR)
    self.RegisterKeyword("attribute", S2MLToken.KEYWORD_ATTRIBUTE)
    self.RegisterKeyword("parameter", S2MLToken.KEYWORD_PARAMETER)
    self.RegisterKeyword("Boolean", S2MLToken.KEYWORD_BOOLEAN)
    self.RegisterKeyword("Integer", S2MLToken.KEYWORD_INTEGER)
    self.RegisterKeyword("Real", S2MLToken.KEYWORD_REAL)
    self.RegisterKeyword("Symbol", S2MLToken.KEYWORD_SYMBOL)
    self.RegisterKeyword("String", S2MLToken.KEYWORD_STRING)
    self.RegisterKeyword("main", S2MLToken.KEYWORD_MAIN)
    self.RegisterKeyword("owner", S2MLToken.KEYWORD_OWNER)
    self.RegisterKeyword("self", S2MLToken.KEYWORD_SELF)
    # Expressions
    # self.RegisterKeyword("add", S2MLToken.KEYWORD_ADD)
    # self.RegisterKeyword("sub", S2MLToken.KEYWORD_SUB)
    # self.RegisterKeyword("mul", S2MLToken.KEYWORD_MUL)
    # self.RegisterKeyword("div", S2MLToken.KEYWORD_DIV)
    self.RegisterKeyword("neg", S2MLToken.KEYWORD_NEG)
    self.RegisterKeyword("min", S2MLToken.KEYWORD_MIN)
    self.RegisterKeyword("max", S2MLToken.KEYWORD_MAX)
    self.RegisterKeyword("pi", S2MLToken.KEYWORD_PI)
    self.RegisterKeyword("false", S2MLToken.KEYWORD_FALSE)
    self.RegisterKeyword("true", S2MLToken.KEYWORD_TRUE)
    self.RegisterKeyword("and", S2MLToken.KEYWORD_AND)
    self.RegisterKeyword("or",  S2MLToken.KEYWORD_OR)
    self.RegisterKeyword("not", S2MLToken.KEYWORD_NOT)
    self.RegisterKeyword("eq",  S2MLToken.KEYWORD_EQ)
    self.RegisterKeyword("df",  S2MLToken.KEYWORD_DF)
    self.RegisterKeyword("lt",  S2MLToken.KEYWORD_LT)
    self.RegisterKeyword("gt",  S2MLToken.KEYWORD_GT)
    self.RegisterKeyword("leq", S2MLToken.KEYWORD_LEQ)
    self.RegisterKeyword("geq", S2MLToken.KEYWORD_GEQ)
    self.RegisterKeyword("within", S2MLToken.KEYWORD_WITHIN)
    self.RegisterKeyword("count", S2MLToken.KEYWORD_COUNT)
    self.RegisterKeyword("if", S2MLToken.KEYWORD_IF)
    self.RegisterKeyword("then", S2MLToken.KEYWORD_THEN)
    self.RegisterKeyword("else", S2MLToken.KEYWORD_ELSE)
    self.RegisterKeyword("uniformDeviate", S2MLToken.KEYWORD_UNIFORM_DEVIATE)
    self.RegisterKeyword("normalDeviate", S2MLToken.KEYWORD_NORMAL_DEVIATE)
    self.RegisterKeyword("lognormalDeviate", S2MLToken.KEYWORD_LOGNORMAL_DEVIATE)
    self.RegisterKeyword("triangularDeviate", S2MLToken.KEYWORD_TRIANGULAR_DEVIATE)
    self.RegisterKeyword("exponentialDeviate", S2MLToken.KEYWORD_EXPONENTIAL_DEVIATE)
    self.RegisterKeyword("WeibullDeviate", S2MLToken.KEYWORD_WEIBULL_DEVIATE)
    self.RegisterKeyword("piecewiseUniformDeviate", S2MLToken.KEYWORD_PIECEWISE_UNIFORM_DEVIATE)
    self.RegisterKeyword("stepwiseDeviate", S2MLToken.KEYWORD_STEPWISE_DEVIATE)
    self.RegisterKeyword("missionTime", S2MLToken.KEYWORD_MISSION_TIME)
    self.RegisterKeyword("exp", S2MLToken.KEYWORD_EXP)
    self.RegisterKeyword("log", S2MLToken.KEYWORD_LOG)
    self.RegisterKeyword("pow", S2MLToken.KEYWORD_POW)
    self.RegisterKeyword("sqrt", S2MLToken.KEYWORD_SQRT)
    self.RegisterKeyword("floor", S2MLToken.KEYWORD_FLOOR)
    self.RegisterKeyword("ceil", S2MLToken.KEYWORD_CEIL)
    self.RegisterKeyword("abs", S2MLToken.KEYWORD_ABS)
    self.RegisterKeyword("mod", S2MLToken.KEYWORD_MOD)
    self.RegisterKeyword("Gamma", S2MLToken.KEYWORD_GAMMA)
    self.RegisterKeyword("sin", S2MLToken.KEYWORD_SIN)
    self.RegisterKeyword("cos", S2MLToken.KEYWORD_COS)
    self.RegisterKeyword("tan", S2MLToken.KEYWORD_TAN)
    self.RegisterKeyword("asin", S2MLToken.KEYWORD_ASIN)
    self.RegisterKeyword("acos", S2MLToken.KEYWORD_ACOS)
    self.RegisterKeyword("atan", S2MLToken.KEYWORD_ATAN)
    self.RegisterKeyword("constantDistribution", S2MLToken.KEYWORD_CONSTANT_DISTRIBUTION)
    self.RegisterKeyword("exponentialDistribution", S2MLToken.KEYWORD_EXPONENTIAL_DISTRIBUTION)
    self.RegisterKeyword("GLMDistribution", S2MLToken.KEYWORD_GLM_DISTRIBUTION)
    self.RegisterKeyword("WeibullDistribution", S2MLToken.KEYWORD_WEIBULL_DISTRIBUTION)
    self.RegisterKeyword("DiracDistribution", S2MLToken.KEYWORD_DIRAC_DISTRIBUTION)
    self.RegisterKeyword("piecewiseUniformDistribution", S2MLToken.KEYWORD_PIECEWISE_UNIFORM_DISTRIBUTION)
    self.RegisterKeyword("stepwiseDistribution", S2MLToken.KEYWORD_STEPWISE_DISTRIBUTION)

  def RegisterKeyword(self, keyword, code):
    self._codesOfKeywords[keyword] = code
    self._stringsOfCodes[code] = keyword

  def GetIdentifierType(self, identifier):
    return self._codesOfKeywords.get(identifier, S2MLToken.TOKEN_IDENTIFIER)

  def GetCodeOfKeyword(self, keyword):
    return self._codesOfKeywords.get(keyword, None)

  def GetStringOfCode(self, code):
    return self._stringsOfCodes.get(code, None)

# 5) Main loop
# ------------

  def Initialize(self, fileName, file):
    self.fileName = fileName
    self.file = file
    self.tokens = []
    self.line = None
    self.lineNumber = 0
    self.columnNumber = 0
    self.endOfFile = 0
    self.tokenLineNumber = 0
    self.tokenColumnNumber = 0
    self.tokenString = None
    self.readChars = []
    self.GetNewLine()
    self.state = S2MLModelLexer.STATE_INITIAL
    return 0

  def Finalize(self):
    pass
     # self.file.close()

  def GetToken(self, tokenIndex):
    if tokenIndex<len(self.tokens):
      return self.tokens[tokenIndex]
    while self.state!=S2MLModelLexer.STATE_END_OF_FILE and tokenIndex>=len(self.tokens):
      if self.state==S2MLModelLexer.STATE_INITIAL:
        c = self.GetChar()
        self.tokenLineNumber = self.lineNumber
        self.tokenColumnNumber = self.columnNumber-1
        if c==None:
          self.state = S2MLModelLexer.STATE_END_OF_FILE
        elif c.isspace():
          continue
        elif c.isalpha() or c=="_":
          self.tokenString = c
          self.state = S2MLModelLexer.STATE_IDENTIFIER
        elif c.isdigit():
          self.tokenString = c
          self.state = S2MLModelLexer.STATE_NUMBER_2
        else:
          action = self.GetAction(c)
          if action==None:
            self.tokenString = c
            self.NewToken(S2MLToken.TOKEN_UNKNOWN)
          else:
            action(self, c)

      # -----------------------------------------------------
      # Identifiers & Paths
      # -----------------------------------------------------
      elif self.state==S2MLModelLexer.STATE_IDENTIFIER:
        c = self.GetChar()
        if c.isdigit() or c.isalpha() or c=="_":
          self.tokenString += c
        elif c==".":
          self.tokenString += c
          self.state = S2MLModelLexer.STATE_PATH_1
        else:
          self.PutCharBack(c)
          type = self.GetIdentifierType(self.tokenString)
          if type==S2MLToken.KEYWORD_MAIN or type==S2MLToken.KEYWORD_OWNER or type==S2MLToken.KEYWORD_SELF:
            self.NewToken(S2MLToken.TOKEN_PATH)
          else:
            self.NewToken(type)
          self.state = S2MLModelLexer.STATE_INITIAL

      elif self.state==S2MLModelLexer.STATE_PATH_1:
        c = self.GetChar()
        if c.isalpha() or c=="_":
          self.tokenString += c
          self.state = S2MLModelLexer.STATE_PATH_2
        else:
          self.NewToken(S2MLToken.TOKEN_UNKNOWN)
          self.state = S2MLModelLexer.STATE_INITIAL
      elif self.state==S2MLModelLexer.STATE_PATH_2:
        c = self.GetChar()
        if c.isdigit() or c.isalpha() or c=="_":
          self.tokenString += c
        elif c==".":
          self.tokenString += c
          self.state = S2MLModelLexer.STATE_PATH_1
        else:
          self.PutCharBack(c)
          self.NewToken(S2MLToken.TOKEN_PATH)
          self.state = S2MLModelLexer.STATE_INITIAL

      # -----------------------------------------------------
      # Numbers
      # -----------------------------------------------------
      elif self.state==S2MLModelLexer.STATE_NUMBER_1:
        c = self.GetChar()
        if c==None:
          self.NewToken(S2MLToken.TOKEN_UNKNOWN)
          self.state = S2MLModelLexer.STATE_END_OF_FILE
        elif c.isdigit():
          self.tokenString += c
          self.state = S2MLModelLexer.STATE_NUMBER_2
        elif c==".":
          self.tokenString += c
          self.state = S2MLModelLexer.STATE_NUMBER_8
        else:
          self.PutCharBack(c)
          self.NewToken(S2MLToken.TOKEN_UNKNOWN)
          self.state = S2MLModelLexer.STATE_INITIAL
      elif self.state==S2MLModelLexer.STATE_NUMBER_2:
        c = self.GetChar()
        if c==None:
          self.NewToken(S2MLToken.TOKEN_INTEGER)
          self.state = S2MLModelLexer.STATE_END_OF_FILE
        elif c.isdigit():
          self.tokenString += c
        elif c==".":
          self.tokenString += c
          self.state = S2MLModelLexer.STATE_NUMBER_3
        elif c=="e" or c=="E":
          self.tokenString += c
          self.state = S2MLModelLexer.STATE_NUMBER_5
        else:
          self.PutCharBack(c)
          self.NewToken(S2MLToken.TOKEN_INTEGER)
          self.state = S2MLModelLexer.STATE_INITIAL
      elif self.state==S2MLModelLexer.STATE_NUMBER_3:
        c = self.GetChar()
        if c==None:
          self.NewToken(S2MLToken.TOKEN_UNKNOWN)
          self.state = S2MLModelLexer.STATE_END_OF_FILE
        elif c.isdigit():
          self.tokenString += c
          self.state = S2MLModelLexer.STATE_NUMBER_4
        elif c=="e" or c=="E":
          self.tokenString += c
          self.state = S2MLModelLexer.STATE_NUMBER_5
        else:
          self.PutCharBack(c)
          self.NewToken(S2MLToken.TOKEN_UNKNOWN)
          self.state = S2MLModelLexer.STATE_INITIAL
      elif self.state==S2MLModelLexer.STATE_NUMBER_4:
        c = self.GetChar()
        if c==None:
          self.NewToken(S2MLToken.TOKEN_REAL)
          self.state = S2MLModelLexer.STATE_END_OF_FILE
        elif c.isdigit():
          self.tokenString += c
        elif c=="e" or c=="E":
          self.tokenString += c
          self.state = S2MLModelLexer.STATE_NUMBER_5
        else:
          self.PutCharBack(c)
          self.NewToken(S2MLToken.TOKEN_REAL)
          self.state = S2MLModelLexer.STATE_INITIAL
      elif self.state==S2MLModelLexer.STATE_NUMBER_5:
        c = self.GetChar()
        if c==None:
          self.NewToken(S2MLToken.TOKEN_UNKNOWN)
          self.state = S2MLModelLexer.STATE_END_OF_FILE
        elif c.isdigit():
          self.tokenString += c
          self.state = S2MLModelLexer.STATE_NUMBER_7
        elif c=="+" or c=="-":
          self.tokenString += c
          self.state = S2MLModelLexer.STATE_NUMBER_6
        else:
          self.PutCharBack(c)
          self.NewToken(S2MLToken.TOKEN_UNKNOWN)
          self.state = S2MLModelLexer.STATE_INITIAL
      elif self.state==S2MLModelLexer.STATE_NUMBER_6:
        c = self.GetChar()
        if c==None:
          self.NewToken(S2MLToken.TOKEN_UNKNOWN)
          self.state = S2MLModelLexer.STATE_END_OF_FILE
        elif c.isdigit():
          self.tokenString += c
          self.state = S2MLModelLexer.STATE_NUMBER_7
        else:
          self.PutCharBack(c)
          self.NewToken(S2MLToken.TOKEN_UNKNOWN)
          self.state = S2MLModelLexer.STATE_INITIAL
      elif self.state==S2MLModelLexer.STATE_NUMBER_7:
        c = self.GetChar()
        if c==None:
          self.NewToken(S2MLToken.TOKEN_REAL)
          self.state = S2MLModelLexer.STATE_END_OF_FILE
        elif c.isdigit():
          self.tokenString += c
        else:
          self.PutCharBack(c)
          self.NewToken(S2MLToken.TOKEN_REAL)
          self.state = S2MLModelLexer.STATE_INITIAL
      elif self.state==S2MLModelLexer.STATE_NUMBER_8:
        c = self.GetChar()
        if c==None:
          self.NewToken(S2MLToken.TOKEN_UNKNOWN)
          self.state = S2MLModelLexer.END_OF_FILE
        elif c.isdigit():
          self.tokenString += c
          self.state = S2MLModelLexer.STATE_NUMBER_4
        else:
          self.PutCharBack(c)
          self.NewToken(S2MLToken.TOKEN_UNKNOWN)
          self.state = S2MLModelLexer.STATE_INITIAL

      # -----------------------------------------------------
      # Single line comments
      # -----------------------------------------------------
      elif self.state==S2MLModelLexer.STATE_SINGLE_LINE_COMMENT:
        c = self.GetChar()
        if c==None:
          if self.keepComments:
            self.NewToken(S2MLToken.TOKEN_COMMENT)
          self.state = 0
        elif c=="\n":
          if self.keepComments:
            self.NewToken(S2MLToken.TOKEN_COMMENT)
          self.state = 1
        else:
          self.tokenString += c

      # -----------------------------------------------------
      # Multiline comments
      # -----------------------------------------------------
      elif self.state==S2MLModelLexer.STATE_MULTILINE_COMMENT_1:
        c = self.GetChar()
        if c==None:
          self.NewToken(S2MLToken.TOKEN_NON_TERMINATED_COMMENT)
          self.state = S2MLModelLexer.STATE_END_OF_FILE
        elif c=="*":
          self.tokenString += c
          self.state = S2MLModelLexer.STATE_MULTILINE_COMMENT_2
        else:
          self.tokenString += c
      elif self.state==S2MLModelLexer.STATE_MULTILINE_COMMENT_2:
        c = self.GetChar()
        if c==None:
          self.NewToken(S2MLToken.TOKEN_NON_TERMINATED_COMMENT)
          self.state = S2MLModelLexer.STATE_END_OF_FILE
        elif c=="*":
          self.tokenString += c
          c = self.GetChar()
        elif c=="/":
          self.tokenString += c
          if self.keepComments:
            self.NewToken(S2MLToken.TOKEN_COMMENT)
          self.state = S2MLModelLexer.STATE_INITIAL
        else:
          self.tokenString += c
          self.state = S2MLModelLexer.STATE_MULTILINE_COMMENT_1

    if tokenIndex>=len(self.tokens):
      return None
    return self.tokens[tokenIndex]

# 6) Auxiliary functions
# ----------------------

  def NewToken(self, type):
    token = S2MLToken(type, self.tokenString, self.fileName, self.tokenLineNumber, self.tokenColumnNumber)
    self.tokens.append(token)

  def GetChar(self):
    if len(self.readChars)!=0:
      return self.readChars.pop()
    if self.columnNumber==len(self.line):
      self.GetNewLine()
      if self.endOfFile:
        return None
      return self.GetChar()
    c = self.line[self.columnNumber]
    self.columnNumber += 1
    return c

  def PutCharBack(self, c):
    self.readChars.append(c)

  def GetNewLine(self):
    self.line = self.file.readline()
    if self.line=="":
      self.endOfFile = 1
      return
    self.lineNumber += 1
    self.columnNumber = 0
    self.line = self.line.rstrip()
    self.line += "\n"

  def ScanFile(self, fileName):
    error = self.Initialize(fileName)
    if error!=0:
      return
    tokenIndex = 0
    while True:
      token = self.GetToken(tokenIndex)
      if token==None:
        break
      tokenIndex += 1
    self.Finalize()

  def PrintTokens(self, outputFile):
    for token in self.tokens:
      outputFile.write(str(token.lineNumber) + "\t")
      outputFile.write(str(token.columnNumber) + "\t")
      outputFile.write(S2MLToken.GetNameOfType(token.type) + "\t")
      outputFile.write(token.string + "\n")
