# S2MLCore version 1.1.1
# S2MLModelPrinter.py
# by Antoine Rauzy
# Copyrights (c) 2018-2022 AltaRica Association
#
# This file is part of S2MLCore.
# S2MLCore is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# S2MLCore is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with S2MLCore.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Models
# 4) Domains, blocks and classes
# 5) Directives
# 6) Elements
# 7) Parameters and stochastic expressions
# 8) Functions and operators
# 9) Matrices
# 10) Auxiliary functions

# 1) Imported modules
# -------------------

import string
import sys
import re
import math

from S2MLMatrices import *
from S2MLStochasticExpression import *
from S2MLModelLexer import *
from S2MLStatistics import *

# 2) Creator
# ----------

class S2MLModelPrinter:
  def __init__(self, engine):
    from S2MLEngine import S2MLEngine
    self._engine = engine
    self._lexer = self._engine.GetModelManager().GetModelLexer()
    self._depth = 0
    self._indentation = "  "
    self._columnSeparator = "\t"
    self._floatFormat = ".3f"

  def RaiseError(self, message):
    self._engine.RaiseError(message)

  def GetIndentation(self):
    return self._indentation

  def SetIndentation(self, indentation):
    self._indentation = indentation

  def GetColumnSeparator(self):
    return self._columnSeparator

  def SetColumnSeparator(self, columnSeparator):
    self._columnSeparator = columnSeparator

  def GetFloatFormat(self):
    return self._floatFormat

  def SetFloatFormat(self, floatFormat):
    self._floatFormat = floatFormat

  def PrintIndentation(self, output):
    for i in range(self._depth):
      output.write(self._indentation)

  def PrintSymbol(self, symbolCode, output):
    output.write(self._lexer.GetStringOfCode(symbolCode))

  def PrintKeyword(self, keywordCode, output):
    output.write(self._lexer.GetStringOfCode(keywordCode))

# 3) Models
# ---------

  def ExportModel(self, model, fileName, mode):
    try:
      output = open(fileName, mode)
    except:
      self.RaiseError('Unable to open file "%s"\n' % fileName)
      return 1
    self.PrintModel(model, output)
    output.close()
    return 0

  def PrintModel(self, model, output):
    self._depth = 0
    for element in model.GetElements():
      self.PrintElement(element, output)

  # Must be derived
  def PrintElement(self, element, output):
    if element.GetType()==S2MLType.DOMAIN_DECLARATION:
      self.PrintDomainDeclaration(element, output)
    elif element.GetType()==S2MLType.BLOCK:
      self.PrintBlock(element, output)
    elif element.GetType()==S2MLType.CLASS:
      self.PrintClass(element, output)
    elif element.GetType()==S2MLType.INSTANCE:
      self.PrintInstance(element, output)
    elif element.GetType()==S2MLType.CLONES:
      self.PrintClonesDirective(element, output)
    elif element.GetType()==S2MLType.EXTENDS:
      self.PrintExtendsDirective(element, output)
    elif element.GetType()==S2MLType.EMBEDS:
      self.PrintEmbedsDirective(element, output)
    elif element.GetType()==S2MLType.ATTRIBUTE:
      self.PrintAttributeDirective(element, output)
    elif element.GetType()==S2MLType.PARAMETER:
      self.PrintParameter(element, output)
    elif element.GetType()==S2MLType.FUNCTION:
      self.PrintFunction(element, output)
    elif element.GetType()==S2MLType.FUNCTION_CALL:
      self.PrintFunctionCall(element, output)
    elif element.GetType()==S2MLType.OPERATOR:
      self.PrintOperator(element, output)

# 4) Domains, blocks and classes
# ------------------------------

  def PrintDomainDeclaration(self, declaration, output):
    self.PrintIndentation(output)
    self.PrintKeyword(S2MLToken.KEYWORD_DOMAIN, output)
    output.write(" ")
    output.write(declaration.GetIdentifier())
    output.write(" ")
    self.PrintDomain(declaration.GetDomain(), output)
    output.write("\n")

  def PrintDomain(self, domain, output):
    if domain.GetType()==S2MLDomain.BOOLEAN:
      self.PrintKeyword(S2MLToken.KEYWORD_BOOLEAN, output)
    elif domain.GetType()==S2MLDomain.INTEGER:
      self.PrintKeyword(S2MLToken.KEYWORD_INTEGER, output)
    elif domain.GetType()==S2MLDomain.REAL:
      self.PrintKeyword(S2MLToken.KEYWORD_REAL, output)
    elif domain.GetType()==S2MLDomain.ENUMERATION:
      self.PrintEnumeration(domain, output)
    elif domain.GetType()==S2MLDomain.RANGE:
      self.PrintRange(domain, output)
    elif domain.GetType()==S2MLDomain.REFERENCE:
     output.write(domain.GetIdentifier())

  def PrintEnumeration(self, domain, output):
    output.write("{")
    first = True
    for constant in domain.GetConstants():
      if first:
        first = False
      else:
        output.write(", ")
      self.PrintConstant(constant, output)
    output.write("}")

  def PrintRange(self, domain, output):
    output.write("[")
    self.PrintConstant(domain.GetMinimumValue(), output)
    output.write(", ")
    self.PrintConstant(domain.GetMaximumValue(), output)
    output.write("]")

  def PrintConstant(self, constant, output, quotedString = True):
    if constant.GetType()==S2MLConstant.NONE:
      pass
    elif constant.GetType()==S2MLConstant.BOOLEAN:
      if constant.GetValue():
        self.PrintKeyword(S2MLToken.KEYWORD_TRUE, output)
      else:
        self.PrintKeyword(S2MLToken.KEYWORD_FALSE, output)
    elif constant.GetType()==S2MLConstant.INTEGER:
      self.PrintInteger(constant.GetValue(), output)
    elif constant.GetType()==S2MLConstant.REAL:
      if constant.GetValue()==math.pi:
        self.PrintKeyword(S2MLToken.KEYWORD_PI, output)
      else:
        self.PrintFloat(constant.GetValue(), output)
    elif constant.GetType()==S2MLConstant.STRING:
      if quotedString:
        output.write('"' + constant.ToString() + '"')
      else:
        output.write(constant.ToString())
    else:
      output.write(constant.GetValue())

  def PrintBlock(self, block, output):
    self.PrintIndentation(output)
    self.PrintKeyword(S2MLToken.KEYWORD_BLOCK, output)
    output.write(" ")
    self.PrintPath(block.GetPath(), output)
    output.write("\n")
    self._PrintBlockBody(block, output)
    self.PrintIndentation(output)
    self.PrintKeyword(S2MLToken.KEYWORD_END, output)
    output.write("\n")

  def PrintClass(self, aClass, output):
    self.PrintIndentation(output)
    self.PrintKeyword(S2MLToken.KEYWORD_CLASS, output)
    output.write(" ")
    self.PrintPath(aClass.GetPath(), output)
    output.write("\n")
    self._PrintBlockBody(aClass, output)
    self.PrintIndentation(output)
    self.PrintKeyword(S2MLToken.KEYWORD_END, output)
    output.write("\n")

  def _PrintBlockBody(self, block, output):
    self._depth += 1
    for element in block.GetElements():
      self.PrintElement(element, output)
    self._depth -= 1

# 5) Directives
# -------------

  def PrintInstance(self, directive, output):
    self.PrintIndentation(output)
    output.write(directive.GetClassIdentifier())
    output.write(" ")
    self.PrintPath(directive.GetPath(), output)
    if directive.GetNumberOfElements()==0:
      output.write(";\n")
    else:
      output.write("\n")
      self._depth += 1
      for element in directive.GetElements():
        self.PrintElement(element, output)
      self._depth -= 1
      self.PrintIndentation(output)
      self.PrintKeyword(S2MLToken.KEYWORD_END, output)
      output.write("\n")

  def PrintClonesDirective(self, directive, output):
    self.PrintIndentation(output)
    self.PrintKeyword(S2MLToken.KEYWORD_CLONES, output)
    output.write(" ")
    self.PrintPath(directive.GetPathToClonedElement(), output)
    output.write(" as ")
    self.PrintPath(directive.GetPath(), output)
    if directive.GetNumberOfElements()==0:
      output.write(";\n")
    else:
      output.write("\n")
      self._depth += 1
      for element in directive.GetElements():
        self.PrintElement(element, output)
      self._depth -= 1
      self.PrintIndentation(output)
      self.PrintKeyword(S2MLToken.KEYWORD_END, output)
      output.write("\n")

  def PrintExtendsDirective(self, directive, output):
    self.PrintIndentation(output)
    self.PrintKeyword(S2MLToken.KEYWORD_EXTENDS, output)
    output.write(" ")
    output.write(directive.GetClassIdentifier())
    output.write(";\n")

  def PrintEmbedsDirective(self, directive, output):
    self.PrintIndentation(output)
    self.PrintKeyword(S2MLToken.KEYWORD_EMBEDS, output)
    output.write(" ")
    self.PrintPath(directive.GetPathToEmbeddedElement(), output)
    output.write(" ")
    self.PrintKeyword(S2MLToken.KEYWORD_AS, output)
    output.write(" ")
    self.PrintPath(directive.GetPath(), output)
    if directive.GetNumberOfElements()==0:
      output.write(";\n")
    else:
      output.write("\n")
      self._depth += 1
      for element in directive.GetElements():
        self.PrintElement(element, output)
      self._depth -= 1
      self.PrintIndentation(output)
      self.PrintKeyword(S2MLToken.KEYWORD_END, output)
      output.write("\n")

  def PrintAttributeDirective(self, directive, output):
    self.PrintIndentation(output)
    self.PrintKeyword(S2MLToken.KEYWORD_ATTRIBUTE, output)
    output.write(" ")
    self.PrintPath(directive.GetPath(), output)
    output.write(" = ")
    self.PrintStochasticExpression(directive.GetValue(), output)
    output.write(";\n")

  def PrintPath(self, path, output):
    output.write(path.ToString())

# 6) Elements
# -----------

  def PrintAttributes(self, element, output):
    if element.GetNumberOfAttributes()==0:
      return
    output.write("(")
    first = True
    for name in element.GetAttributeNames():
      if first:
        first = False
      else:
        output.write(", ")
      output.write(name)
      output.write("=")
      value = element.GetAttribute(name)
      self.PrintStochasticExpression(value, output)
    output.write(")")

# 7) Parameters and stochastic expressions
# ----------------------------------------

  def PrintParameter(self, parameter, output):
    self.PrintIndentation(output)
    self.PrintKeyword(S2MLToken.KEYWORD_PARAMETER, output)
    output.write(" ")
    self.PrintDomain(parameter.GetDomain(), output)
    output.write(" ")
    self.PrintPath(parameter.GetPath(), output)
    output.write(" = ")
    self.PrintStochasticExpression(parameter.GetDefinition(), output)
    output.write(";\n")

  def PrintStochasticExpression(self, expression, output):
    if expression.GetType()==S2MLType.REFERENCE_EXPRESSION:
      self.PrintPath(expression.GetReferencePath(), output)
    elif expression.GetType()==S2MLType.CONSTANT_EXPRESSION:
      self.PrintConstant(expression.GetValue(), output)
    elif expression.GetType()==S2MLType.AND_EXPRESSION:
      self.PrintChildrenOfStochasticExpression(expression, " and ", output)
    elif expression.GetType()==S2MLType.OR_EXPRESSION:
      self.PrintChildrenOfStochasticExpression(expression, " or ", output)
    elif expression.GetType()==S2MLType.NOT_EXPRESSION:
      output.write("not ")
      child = expression.GetChild(0)
      self.PrintStochasticExpression(child, output)
    elif expression.GetType()==S2MLType.EQ_EXPRESSION:
      self.PrintChildrenOfStochasticExpression(expression, " == ", output)
    elif expression.GetType()==S2MLType.DF_EXPRESSION:
      self.PrintChildrenOfStochasticExpression(expression, " != ", output)
    elif expression.GetType()==S2MLType.LT_EXPRESSION:
      self.PrintChildrenOfStochasticExpression(expression, " < ", output)
    elif expression.GetType()==S2MLType.GT_EXPRESSION:
      self.PrintChildrenOfStochasticExpression(expression, " > ", output)
    elif expression.GetType()==S2MLType.LEQ_EXPRESSION:
      self.PrintChildrenOfStochasticExpression(expression, " <= ", output)
    elif expression.GetType()==S2MLType.GEQ_EXPRESSION:
      self.PrintChildrenOfStochasticExpression(expression, " >= ", output)
    elif expression.GetType()==S2MLType.NEG_EXPRESSION:
      output.write("-")
      child = expression.GetChild(0)
      self.PrintStochasticExpression(child, output)
    elif expression.GetType()==S2MLType.ADD_EXPRESSION:
      self.PrintChildrenOfStochasticExpression(expression, " + ", output)
    elif expression.GetType()==S2MLType.SUB_EXPRESSION:
      self.PrintChildrenOfStochasticExpression(expression, " - ", output)
    elif expression.GetType()==S2MLType.MUL_EXPRESSION:
      self.PrintChildrenOfStochasticExpression(expression, " * ", output)
    elif expression.GetType()==S2MLType.DIV_EXPRESSION:
      self.PrintChildrenOfStochasticExpression(expression, " / ", output)
    elif expression.GetType()==S2MLType.IF_EXPRESSION:
      self.PrintKeyword(S2MLToken.KEYWORD_IF, output)
      output.write(" ")
      child = expression.GetChild(0)
      self.PrintStochasticExpression(child, output)
      output.write(" ")
      self.PrintKeyword(S2MLToken.KEYWORD_THEN, output)
      output.write(" ")
      child = expression.GetChild(1)
      self.PrintStochasticExpression(child, output)
      output.write(" ")
      self.PrintKeyword(S2MLToken.KEYWORD_ELSE, output)
      output.write(" ")
      child = expression.GetChild(2)
      self.PrintStochasticExpression(child, output)
    elif expression.GetType()==S2MLType.WITHIN_EXPRESSION:
      self.PrintKeyword(S2MLToken.KEYWORD_WITHIN, output)
      output.write(" ")
      child = expression.GetChild(0)
      self.PrintStochasticExpression(child, output)
      output.write(" ")
      domain = expression.GetDomain()
      self.PrintDomain(domain, output)
    elif expression.GetType()==S2MLType.OPERATOR_CALL_EXPRESSION:
      output.write(expression.GetFunctor())
      output.write("(")
      self.PrintChildrenOfStochasticExpression(expression, ", ", output)
      output.write(")")
    elif expression.IsEmpiricalDistribution():
      operator = self._engine.GetExpressionManager().GetOperatorOfType(expression.GetType())
      keywordCode = operator.GetKeywordCode()
      self.PrintKeyword(keywordCode, output)
      output.write(" ")
      distribution = expression.GetDistribution()
      self.PrintDistribution(distribution, output)
      output.write("(")
      self.PrintChildrenOfStochasticExpression(expression, ", ", output)
      output.write(")")
    else:
      operator = self._engine.GetExpressionManager().GetOperatorOfType(expression.GetType())
      keywordCode = operator.GetKeywordCode()
      self.PrintKeyword(keywordCode, output)
      output.write("(")
      self.PrintChildrenOfStochasticExpression(expression, ", ", output)
      output.write(")")

  def PrintChildrenOfStochasticExpression(self, expression, separator, output):
    first = True
    for child in expression.GetChildren():
      if first:
        first = False
      else:
        output.write(separator)
      if child.GetPriority()>=expression.GetPriority() and child.GetPriority()>0:
        output.write("(")
      self.PrintStochasticExpression(child, output)
      if child.GetPriority()>=expression.GetPriority() and child.GetPriority()>0:
        output.write(")")

  def PrintDistribution(self, distribution, output):
    output.write("[")
    first = True
    for point in distribution.GetPoints():
      if first:
        first = False
      else:
        output.write(", ")
      output.write("(")
      output.write("%g" % point.GetX())
      output.write(", ")
      output.write("%g" % point.GetY())
      output.write(")")
    output.write("]")

# 8) Functions and operators
# --------------------------

  def PrintFunction(self, function, output):
    self.PrintIndentation(output)
    self.PrintKeyword(S2MLToken.KEYWORD_FUNCTION, output)
    output.write(" ")
    output.write(function.GetIdentifier())
    for argument in function.GetArguments():
      output.write(" ")
      output.write(argument)
    output.write(" =")
    output.write("\n")
    self._depth += 1
    self._PrintBlockBody(function, output)
    self._depth -= 1
    self.PrintIndentation(output)
    self.PrintKeyword(S2MLToken.KEYWORD_END, output)
    output.write("\n")

  def PrintFunctionCall(self, directive, output):
    self.PrintIndentation(output)
    output.write("(")
    output.write(directive.GetFunctionIdentifier())
    output.write(" ")
    first = True
    for argument in directive.GetArguments():
      output.write(" ")
      self.PrintStochasticExpression(argument, output)
    output.write(");\n")

  def PrintOperator(self, operator, output):
    self.PrintIndentation(output)
    self.PrintKeyword(S2MLToken.KEYWORD_OPERATOR, output)
    output.write(" ")
    output.write(operator.GetIdentifier())
    for argument in operator.GetArguments():
      output.write(" ")
      output.write(argument)
    output.write(" =")
    output.write("\n")
    self._depth += 1
    self.PrintIndentation(output)
    self.PrintStochasticExpression(operator.GetDefinition(), output)
    self._depth -= 1
    output.write(";")
    output.write("\n")

# 9) Matrices
# -----------

  def ExportMatrix(self, matrix, fileName, mode="w"):
    try:
      output = open(fileName, mode)
    except:
      self.RaiseError('Unable to open file "%s"\n' % fileName)
      return 1
    self.PrintMatrix(matrix, output)
    output.close()
    return 0

  def PrintMatrix(self, matrix, output):
    for rowIndex in range(matrix.GetNumberOfRows()):
      rowVector = matrix.GetRow(rowIndex)
      self.PrintVector(rowVector, output)
      output.write("\n")
    output.write("\n")
    output.flush()

  def PrintVector(self, vector, output):
    for index in range(vector.GetSize()):
     if index>0:
       output.write(self._columnSeparator)
     constant = vector.Get(index)
     self.PrintConstant(constant, output, False)

# 10) Auxiliary functions
# -----------------------

  def PrintBoolean(self, value, output):
    if value:
      output.write("true")
    else:
      output.write("false")

  def PrintInteger(self, value, output):
    output.write(str(value))

  def PrintFloat(self, value, output):
    output.write("{1:{0}}".format(self._floatFormat, value))

