# S2MLCore version 1.1.1
# S2MLModel.py
# by Antoine Rauzy
# Copyrights (c) 2018-2022 AltaRica Association
#
# This file is part of S2MLCore.
# S2MLCore is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# S2MLCore is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with S2MLCore.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Management of constants and domains
# 4) Type testing
# 5) Element factory (to be derived)
# 6) Management of classes
# 7) Tests for the parser

# 1) Imported modules
# -------------------

from S2MLDomains import *
from S2MLMetaModel import *

from S2MLType import *
from S2MLStochasticExpression import *
from S2MLBaseBlock import *
from S2MLDomainDeclaration import *
from S2MLParameter import *
from S2MLOperator import *
from S2MLFunction import *
from S2MLFunctionCall import *
from S2MLBlock import *
from S2MLClass import *
from S2MLInstance import *
from S2MLExtendsDirective import *
from S2MLClonesDirective import *
from S2MLEmbedsDirective import *
from S2MLAttributeDirective import *

# 2) Creator
# ----------

class S2MLModel(S2MLBaseBlock):

  def __init__(self):
    S2MLBaseBlock.__init__(self, S2MLType.MODEL, None)
    self.SetPathAndIdentifierAndInsertInContainerDictionary("__MODEL__")
    self._universe = S2MLUniverse()

# 3) Management of constants and domains
# --------------------------------------

  def GetNone(self):
    return self._universe.GetNone()
  def NewNone(self):
    return self._universe.NewNone()

  def GetBoolean(self, value):
    return self._universe.GetBoolean(value)
  def NewBoolean(self):
    return self._universe.NewBoolean(value)

  def GetInteger(self, value):
    return self._universe.GetInteger(value)
  def NewInteger(self):
    return self._universe.NewInteger(value)

  def GetReal(self, value):
    return self._universe.GetReal(value)
  def NewReal(self):
    return self._universe.NewReal(value)

  def GetSymbol(self, value):
    return self._universe.GetSymbol(value)
  def NewSymbol(self, value):
    return self._universe.NewSymbol(value)

  def GetString(self, value):
    return self._universe.GetString(value)
  def NewString(self, value):
    return self._universe.NewString(value)

  def GetConstant(self, constant):
    return self._universe.GetConstant(constant)
  def NewConstant(self, constant):
    return self._universe.NewConstant(constant)

  def NewBooleanDomain(self):
    return S2MLDomain.NewBoolean(self._universe)

  def NewIntegerDomain(self):
    return S2MLDomain.NewInteger(self._universe)

  def NewRealDomain(self):
    return S2MLDomain.NewReal(self._universe)

  def NewSymbolDomain(self):
    return S2MLDomain.NewSymbol(self._universe)

  def NewEnumerationDomain(self):
    return S2MLDomain.NewEnumeration(self._universe)

  def NewRangeDomain(self, lowerBound, upperBound):
    return S2MLDomain.NewRange(self._universe, lowerBound, upperBound)

  def NewReferenceDomain(self, identifier):
    return S2MLDomain.NewReference(self._universe, identifier)

# 4) Type testing
# ---------------

  @staticmethod
  def IsDomainDeclaration(element):
    return element.GetType()==S2MLType.DOMAIN_DECLARATION
  @staticmethod
  def IsBlock(element):
    return element.GetType()==S2MLType.BLOCK
  @staticmethod
  def IsClass(element):
    return element.GetType()==S2MLType.CLASS
  @staticmethod
  def IsFunction(element):
    return element.GetType()==S2MLType.FUNCTION
  @staticmethod
  def IsOperator(element):
    return element.GetType()==S2MLType.OPERATOR

  @staticmethod
  def IsParameter(element):
    return element.GetType()==S2MLType.PARAMETER

  @staticmethod
  def IsModel(element):
    return element.GetType()==S2MLType.MODEL

  def IsValidReferredElementInStochasticExpression(self, element):
    return element.GetType()==S2MLType.PARAMETER

# 5) Element factory (to be derived)
# ----------------------------------

  def NewDomainDeclaration(self, identifier, domain, container):
    return S2MLDomainDeclaration(identifier, domain, container)

  def NewPort(self, type, domain, path, container):
    return S2MLPort(type, domain, path, container)

  def NewParameter(self, domain, path, container):
    return S2MLParameter(domain, path, container)

  def NewBlock(self, path, container):
    return S2MLBlock(path, container)

  def NewClass(self, identifier, container):
    return S2MLClass(identifier, container)

  def NewInstance(self, identifier, classIdentifier, container):
    return S2MLInstance(identifier, classIdentifier, container)

  def NewExtendsDirective(self, classIdentifier, container):
    return S2MLExtendsDirective(classIdentifier, container)

  def NewClonesDirective(self, identifier, pathToClonedElement, container):
    return S2MLClonesDirective(identifier, pathToClonedElement, container)

  def NewEmbedsDirective(self, identifier, pathToEmbeddedElement, container):
    return S2MLEmbedsDirective(identifier, pathToEmbeddedElement, container)

  def NewAttributeDirective(self, path, expression, container):
    return S2MLAttributeDirective(path, expression, container)

  def NewFunction(self, identifier, container):
    return S2MLFunction(identifier, container)

  def NewFunctionCall(self, functionIdentifier, container):
    return S2MLFunctionCall(functionIdentifier, container)

  def NewOperator(self, identifier, container):
    return S2MLOperator(identifier, container)

# 6) Management of classes
# ------------------------

  def LookForClass(self, identifier):
    element = self.LookForIdentifiedElement(identifier)
    if element==None or not S2MLModel.IsClass(element):
      return None
    return element

  def CollectClasses(self):
    classes = []
    self.CollectElements(S2MLModel.IsClass, False, classes)
    return classes

  def GetNumberOfClasses(self):
    return len(self.CollectClasses())

  def LookForFunction(self, identifier):
    element = self.LookForIdentifiedElement(identifier)
    if element==None or not S2MLModel.IsFunction(element):
      return None
    return element

  def LookForOperator(self, identifier):
    element = self.LookForIdentifiedElement(identifier)
    if element==None or not S2MLModel.IsOperator(element):
      return None
    return element

# 7) Tests for the parser
# -----------------------

  def CanComposeDomain(self, container):
    return S2MLModel.IsModel(container)

  def CanComposePort(self, container):
    return True

  def CanComposeBlock(self, container):
    return True

  def CanComposeClass(self, container):
    return S2MLModel.IsModel(container)

  def CanComposeInstance(self, container):
    return True

  def CanComposeClonesDirective(self, container):
    return S2MLModel.IsBlock(container) or S2MLModel.IsClass(container)

  def CanComposeExtendsDirective(self, container):
    return S2MLModel.IsBlock(container) or S2MLModel.IsClass(container)

  def CanComposeEmbedsDirective(self, container):
    return S2MLModel.IsBlock(container) or S2MLModel.IsClass(container)

  def CanComposeAttributeDirective(self, container):
    return True

  def CanComposeParameter(self, container):
    return True

  def CanComposeFunction(self, container):
    return S2MLModel.IsModel(container)

  def CanComposeFunctionCall(self, container):
    return S2MLModel.IsBlock(container) or S2MLModel.IsClass(container)

  def CanComposeOperator(self, container):
    return S2MLModel.IsModel(container)




