# S2MLCore version 1.1.1
# S2MLModelFlattener.py
# by Antoine Rauzy
# Copyrights (c) 2018-2022 AltaRica Association
#
# This file is part of S2MLCore, a pedagogical calculation engine for Boolean reliability models.
# S2MLCore is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# S2MLCore is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with S2MLCore.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Actions
# 3) Flattener
#   3.1) Creator
#   3.2) Error management
#   3.3) Compilation loop
#   3.4) Instantiation of models
#   3.5) Instantiation of directives
#   3.6) Instantiation of elements
#   3.7) Flattening of models

# 1) Imported modules
# -------------------

from S2MLMetaModel import *
from S2MLModels import *
from S2MLInterface import *

# 2) Aliases and Environments
# ---------------------------

# 2.1) Aliases
# ------------

class S2MLFlattenerAlias:
  def __init__(self, identifier, expression):
    self._identifier = identifier
    self._expression = expression

  def GetIdentifier(self):
    return self._identifier

  def SetIdentifier(self, identifier):
    self._identifier = identifier

  def GetExpression(self):
    return self._expression

  def SetExpression(self, expression):
    self._expression = expression

# 2.2) Environments
# -----------------

class S2MLFlattenerEnvironment:
  def __init__(self):
    self._aliases = dict()

  def GetValueOfIdentifierAsExpression(self, identifier):
    alias = self._aliases.get(identifier, None)
    if alias!=None:
      return alias.GetExpression()
    if self._parent!=None:
      return self._parent.GetValueOfIdentifierAsExpression(identifier)
    return None

  def GetValueOfIdentifierAsPath(self, identifier):
    value = self.GetValueOfIdentifierAsExpression()
    if value==None:
      return None
    if value.GetType()==S2MLType.REFERENCE_EXPRESSION:
      return value.GetReferencePath()
    return None

  def SetValueOfIdentifier(self, identifier, expression):
    self._aliases[identifier] = S2MLFlattenerAlias(identifier, expression)

# 2) Actions
# ----------

class S2MLFlattenerAction:
  INSTANTIATE_ELEMENT = 1
  FLATTEN_ELEMENT = 2
  RESTORE_ENVIRONEMENT = 3

  TODO = 1
  POSTPONED = 2
  COMPLETED = 3
  FAILED = 4

  def __init__(self, type, size):
    self._type = type
    self._status = S2MLFlattenerAction.TODO
    self._size = size
    self._data = [None for _ in range(0, size)]

  def GetType(self):
    return self._type

  def GetStatus(self):
    return self._status

  def SetStatus(self, status):
    self._status = status

  def GetSize(self):
    return self._size

  def GetDatum(self, index):
    if index<0 or index>=self.GetSize():
      return None
    return self._data[index]

  def SetDatum(self, index, datum):
    if index<0 or index>=self.GetSize():
      return
    self._data[index] = datum

  @staticmethod
  def NewInstantiateElement(sourceElement, targetContainer):
    self = S2MLFlattenerAction(S2MLFlattenerAction.INSTANTIATE_ELEMENT, 2)
    self.SetDatum(0, sourceElement)
    self.SetDatum(1, targetContainer)
    return self

  @staticmethod
  def NewFlattenElement(sourceContainer, sourceElement, targetContainer):
    self = S2MLFlattenerAction(S2MLFlattenerAction.FLATTEN_ELEMENT, 3)
    self.SetDatum(0, sourceContainer)
    self.SetDatum(1, sourceElement)
    self.SetDatum(2, targetContainer)
    return self

  @staticmethod
  def NewRestoreEnvironment(previousEnvironment):
    self = S2MLFlattenerAction(S2MLFlattenerAction.RESTORE_ENVIRONEMENT, 1)
    self.SetDatum(0, previousEnvironment)
    return self

# 3) Flattener
# ------------

# 3.1) Creator
# ------------

class S2MLModelFlattener:
  def __init__(self, engine):
    from S2MLEngine import S2MLEngine
    self._engine = engine
    self._printer = self._engine.GetModelManager().GetModelPrinter()
    self._sourceModel = None
    self._targetModel = None
    self._actionStack = []
    self._environment = None
    self._errors = 0
    self._noPostponing = False

# 3.2) Error management
# ---------------------

  def GetNumberOfErrors(self):
    return self._errors

  def RaiseError(self, message):
    fullMessage = ""
    self._errors += 1
    if self._errors>1:
      fullMessage += "\n"
    fullMessage += "Error %d\n" % self._errors
    fullMessage += message
    self._engine.RaiseError(message)

# 3.3) Compilation loop
# ---------------------

  def GetEnvironment(self):
    return self._environment

  def GetValueOfIdentifierAsExpression(self, identifier):
    if self._environment==None:
      return None
    return self._environment.GetValueOfIdentifierAsExpression(identifier)

  def GetValueOfIdentifierAsPath(self, identifier):
    if self._environment==None:
      return None
    return self._environment.GetValueOfIdentifierAsPath(identifier)

  def _IsActionStackEmpty(self):
    return len(self._actionStack)==0

  def _PushAction(self, action):
    self._actionStack.insert(0, action)

  def _EnqueueAction(self, action):
    self._actionStack.append(action)

  def _DequeueStack(self):
    return self._actionStack.pop(0)

  def _ExecuteCompilationLoop(self):
    firstPostponedAction = None
    self._noPostponing = False
    while not self._IsActionStackEmpty():
      action = self._DequeueStack()
      status = self._PerformAction(action)
      if status==S2MLFlattenerAction.COMPLETED:
        firstPostponedAction = None
      elif status==S2MLFlattenerAction.POSTPONED:
        self._EnqueueAction(action)
        if action==firstPostponedAction:
          break
        else:
          action.SetStatus(S2MLFlattenerAction.POSTPONED)
          if firstPostponedAction==None:
            firstPostponedAction = action
      elif status==S2MLFlattenerAction.FAILED:
        break
    if self._errors>0:
      while not self._IsActionStackEmpty():
        action = self._DequeueStack()
      return
    self._noPostponing = True
    while not self._IsActionStackEmpty():
      action = self._DequeueStack()
      status = self._PerformAction(action)
      if status==S2MLFlattenerAction.FAILED:
        break

  def _PerformAction(self, action):
    if action.GetType()==S2MLFlattenerAction.INSTANTIATE_ELEMENT:
      return self.InstantiateElement(action)
    elif action.GetType()==S2MLFlattenerAction.FLATTEN_ELEMENT:
      return self.FlattenElement(action)
    elif action.GetType()==S2MLFlattenerAction.RESTORE_ENVIRONEMENT:
      previousEnvironment = action.GetDatum(0)
      self._environment = previousEnvironment
      return S2MLFlattenerAction.COMPLETED
    return S2MLFlattenerAction.FAILED

# 3.4) Instantiation of models
# ----------------------------

  def InstantiateModel(self, sourceModel):
    from S2MLEngine import S2MLEngine
    self._sourceModel = sourceModel
    self._targetModel = self._engine.GetModelManager().NewModel()
    self._errors = 0
    action = S2MLFlattenerAction.NewInstantiateElement(self._sourceModel, self._targetModel)
    self._PushAction(action)
    self._ExecuteCompilationLoop()
    if self._errors!=0:
      return None
    return self._targetModel

  def InstantiateAndFlattenModel(self, sourceModel):
    from S2MLEngine import S2MLEngine
    self._sourceModel = sourceModel
    self._targetModel = self._engine.GetModelManager().NewModel()
    self._errors = 0
    action = S2MLFlattenerAction.NewInstantiateElement(self._sourceModel, self._targetModel)
    self._PushAction(action)
    self._ExecuteCompilationLoop()
    if self._errors!=0:
      return None
    self._sourceModel = self._targetModel
    self._targetModel = self._engine.NewModel()
    action = S2MLFlattenerAction.NewFlattenElement(self._sourceModel, self._sourceModel, self._targetModel)
    self._PushAction(action)
    self._ExecuteCompilationLoop()
    if self._errors!=0:
      return None
    return self._targetModel

  def InstantiateElement(self, action):
    sourceElement = action.GetDatum(0)
    if sourceElement.GetType()==S2MLType.MODEL:
      status = self._InstantiateModel(action)
    elif sourceElement.GetType()==S2MLType.DOMAIN_DECLARATION:
      status = self._InstantiateDomainDeclaration(action)
    elif sourceElement.GetType()==S2MLType.BLOCK:
      status = self._InstantiateBlock(action)
    elif sourceElement.GetType()==S2MLType.CLASS:
      status = S2MLFlattenerAction.COMPLETED
    elif sourceElement.GetType()==S2MLType.INSTANCE:
      status = self._InstantiateInstance(action)
    elif sourceElement.GetType()==S2MLType.CLONES:
      status = self._InstantiateClonesDirective(action)
    elif sourceElement.GetType()==S2MLType.EXTENDS:
      status = self._InstantiateExtendsDirective(action)
    elif sourceElement.GetType()==S2MLType.EMBEDS:
      status = self._InstantiateEmbedsDirective(action)
    elif sourceElement.GetType()==S2MLType.ATTRIBUTE:
      status = self._InstantiateAttributeDirective(action)
    elif sourceElement.GetType()==S2MLType.PARAMETER:
      status = self._InstantiateParameter(action)
    elif sourceElement.GetType()==S2MLType.FUNCTION:
      status = S2MLFlattenerAction.COMPLETED
    elif sourceElement.GetType()==S2MLType.FUNCTION_CALL:
      status = self._InstantiateFunctionCall(action)
    elif sourceElement.GetType()==S2MLType.OPERATOR:
      status = S2MLFlattenerAction.COMPLETED
    else:
      status = S2MLFlattenerAction.FAILED
    return status

  def _PushInstantiateActionsForElements(self, sourceContainer, targetContainer):
    for sourceElement in reversed(sourceContainer.GetElements()):
      self._PushInstantiateActionForElement(sourceElement, targetContainer)

  def _PushInstantiateActionForElement(self, sourceElement, targetContainer):
    action = S2MLFlattenerAction.NewInstantiateElement(sourceElement, targetContainer)
    self._PushAction(action)

  def _InstantiateModel(self, action):
    sourceModel = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    self._PushInstantiateActionsForElements(sourceModel, targetContainer)
    return S2MLFlattenerAction.COMPLETED

  def _InstantiateDomainDeclaration(self, action):
    sourceDeclaration = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    sourceDomain = sourceDeclaration.GetDomain()
    targetDomain = self._CloneDomain(sourceDomain)
    targetDeclaration = self._targetModel.NewDomainDeclaration(sourceDeclaration.GetIdentifier(), targetDomain, targetContainer)
    return S2MLFlattenerAction.COMPLETED

  def _CloneDomain(self, sourceDomain):
    if sourceDomain.GetType()==S2MLDomain.BOOLEAN:
      targetDomain = self._targetModel.NewBooleanDomain()
    elif sourceDomain.GetType()==S2MLDomain.INTEGER:
      targetDomain = self._targetModel.NewIntegerDomain()
    elif sourceDomain.GetType()==S2MLDomain.REAL:
      targetDomain = self._targetModel.NewRealDomain()
    elif sourceDomain.GetType()==S2MLDomain.SYMBOL:
      targetDomain = self._targetModel.NewSymbolDomain()
    elif sourceDomain.GetType()==S2MLDomain.ENUMERATION:
      targetDomain = self._targetModel.NewEnumerationDomain()
      for sourceConstant in sourceDomain.GetConstants():
        targetConstant = self._targetModel.NewConstant(sourceConstant)
        targetDomain.AppendConstant(targetConstant)
    elif sourceDomain.GetType()==S2MLDomain.RANGE:
      lowerBound = sourceDomain.GetLowerBound()
      upperBound = sourceDomain.GetUpperBound()
      targetDomain = self._targetModel.NewRangeDomain(lowerBound, upperBound)
    elif sourceDomain.GetType()==S2MLDomain.REFERENCE:
      identifier = sourceDomain.GetIdentifier()
      targetDomain = self._targetModel.NewReferenceDomain(identifier)
    return targetDomain

  def _InstantiateBlock(self, action):
    sourceBlock = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    actualTargetContainer = self._GetActualTargetContainer(sourceBlock, targetContainer)
    if actualTargetContainer==None:
      return S2MLFlattenerAction.FAILED
    sourcePath = sourceBlock.GetPath()
    targetIdentifier = sourcePath.GetLastIdentifier()
    targetBlock = actualTargetContainer.LookForIdentifiedElement(targetIdentifier)
    if targetBlock==None:
      targetPath = S2MLPath.NewEmptyPath()
      targetPath.AppendIdentifier(targetIdentifier)
      targetBlock = self._targetModel.NewBlock(targetPath, actualTargetContainer)
    elif targetBlock.GetType()!=S2MLType.BLOCK:
      absoluteSourcePath = sourceBlock.GetAbsolutePath()
      self.RaiseError("Type mismatch for block %s\n" % absoluteSourcePath.ToString())
      return S2MLFlattenerAction.FAILED
    self._PushInstantiateActionsForElements(sourceBlock, targetBlock)
    return S2MLFlattenerAction.COMPLETED

# 3.5) Instantiation of directives
# --------------------------------

  def _InstantiateInstance(self, action):
    sourceDirective = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    sourceClassIdentifier = sourceDirective.GetClassIdentifier()
    sourceClass = self._sourceModel.LookForClass(sourceClassIdentifier)
    if sourceClass==None:
      parentBlock = sourceDirective.GetContainer()
      absoluteSourcePath = parentBlock.GetAbsolutePath()
      message = "In {0:s}, attempt to instantiate undefined class {1:s}\n".format(absoluteSourcePath.ToString(), sourceClassIdentifier)
      self.RaiseError(message)
      return S2MLFlattenerAction.FAILED
    targetIdentifier = sourceDirective.GetIdentifier()
    targetBlock = targetContainer.LookForIdentifiedElement(targetIdentifier)
    if targetBlock!=None:
      parentBlock = sourceDirective.GetContainer()
      absoluteSourcePath = parentBlock.GetAbsolutePath()
      message = "In {0:s}, element {1:s} already exists\n".format(absoluteSourcePath.ToString(), targetIdentifier)
      self.RaiseError(message)
      return S2MLFlattenerAction.FAILED
    targetPath = S2MLPath.NewEmptyPath()
    targetPath.AppendIdentifier(targetIdentifier)
    targetBlock = self._targetModel.NewBlock(targetPath, targetContainer)
    self._PushInstantiateActionsForElements(sourceDirective, targetBlock)
    self._PushInstantiateActionsForElements(sourceClass, targetBlock)
    return S2MLFlattenerAction.COMPLETED

  def _InstantiateExtendsDirective(self, action):
    sourceDirective = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    sourceClassIdentifier = sourceDirective.GetClassIdentifier()
    sourceClass = self._sourceModel.LookForClass(sourceClassIdentifier)
    if sourceClass==None:
      message = "In {0:s}, attempt to inherit undefined class {1:s}\n".format(sourceContainerAbsolutePath.ToString(), sourceClassIdentifier)
      self.RaiseError(message)
      return S2MLFlattenerAction.FAILED
    self._PushInstantiateActionsForElements(sourceClass, targetContainer)
    return S2MLFlattenerAction.COMPLETED

  def _InstantiateClonesDirective(self, action):
    sourceDirective = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    pathToClonedElement = sourceDirective.GetPathToClonedElement()
    sourceBlock = targetContainer.AccessNamedElement(pathToClonedElement)
    if sourceBlock==None:
      absoluteSourcePath = sourceDirective.GetAbsolutePath()
      message = "In {0:s}, attempt to clone undefined block {1:s}\n".format(absoluteSourcePath.ToString(), pathToClonedElement.ToString())
      self.RaiseError(message)
      return S2MLFlattenerAction.FAILED
    targetIdentifier = sourceDirective.GetIdentifier()
    targetBlock = targetContainer.LookForIdentifiedElement(targetIdentifier)
    if targetBlock!=None:
      absoluteSourcePath = sourceDirective.GetAbsolutePath()
      message = "In {0:s}, element {1:s} already exists\n".format(absoluteSourcePath.ToString(), targetIdentifier)
      self.RaiseError(message)
      return S2MLFlattenerAction.FAILED
    targetPath = S2MLPath.NewEmptyPath()
    targetPath.AppendIdentifier(targetIdentifier)
    targetBlock = self._targetModel.NewBlock(targetPath, targetContainer)
    self._PushInstantiateActionsForElements(sourceDirective, targetBlock)
    self._PushInstantiateActionsForElements(sourceBlock, targetBlock)
    return S2MLFlattenerAction.COMPLETED

  def _InstantiateEmbedsDirective(self, action):
    sourceDirective = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetIdentifier = sourceDirective.GetIdentifier()
    targetBlock = targetContainer.LookForBlock(targetIdentifier)
    if targetBlock!=None:
      absoluteSourcePath = sourceDirective.GetAbsolutePath()
      message = "In {0:s}, element {1:s} already exists\n".format(absoluteSourcePath.ToString(), targetIdentifier)
      self.RaiseError(message)
    pathToEmbeddedElement = sourceDirective.GetPathToEmbeddedElement()
    targetContainer.SetAlias(targetIdentifier, pathToEmbeddedElement)
    return S2MLFlattenerAction.COMPLETED

  def _InstantiateAttributeDirective(self, action):
    sourceDirective = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    sourcePath = sourceDirective.GetPath()
    targetIdentifier = sourcePath.GetLastIdentifier()
    pathToTargetElement = S2MLPath.Copy(sourcePath)
    pathToTargetElement.RemoveLastIdentifier()
    targetElement = targetContainer.AccessNamedElement(pathToTargetElement)
    if targetElement==None:
      absoluteSourcePath = sourceDirective.GetAbsolutePath()
      self.RaiseError("Attribute directive {0:s} refers to no element\n".format(absoluteSourcePath.ToString()))
      return S2MLFlattenerAction.FAILED
    actualTargetContainer = targetElement.GetContainer()
    sourceExpression = sourceDirective.GetValue()
    targetExpression = self._ContextualizedCopyOfStochasticExpression(sourceExpression, targetContainer, actualTargetContainer)
    if targetExpression==None:
      return S2MLFlattenerAction.FAILED
    targetElement.SetAttribute(targetIdentifier, targetExpression)
    return S2MLFlattenerAction.COMPLETED

  def _InstantiateFunctionCall(self, action):
    sourceDirective = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    actualTargetContainer = self._GetActualTargetContainer(sourceDirective, targetContainer)
    sourceFunctionIdentifier = sourceDirective.GetFunctionIdentifier()
    sourceFunction = self._sourceModel.LookForFunction(sourceFunctionIdentifier)
    if sourceFunction==None:
      message = "In {0:s}, attempt to apply undefined rule {1:s}\n".format(sourceContainerAbsolutePath.ToString(), sourceFunctionIdentifier)
      self.RaiseError(message)
      return S2MLFlattenerAction.FAILED
    if sourceDirective.GetNumberOfArguments()!=sourceFunction.GetNumberOfArguments():
      message = "In {0:s}, wrong number of arguments in application of rule {1:s}\n".format(sourceContainerAbsolutePath.ToString(), sourceFunctionIdentifier)
      self.RaiseError(message)
      return S2MLFlattenerAction.FAILED
    newEnvironment = S2MLFlattenerEnvironment()
    for index in range(0, sourceDirective.GetNumberOfArguments()):
      formalArgument = sourceFunction.GetArgument(index)
      actualArgument = sourceDirective.GetArgument(index)
      targetArgument = self._ContextualizedCopyOfStochasticExpression(actualArgument, targetContainer, actualTargetContainer)
      if targetArgument==None:
        return None
      newEnvironment.SetValueOfIdentifier(formalArgument, targetArgument)
    previousEnvironment = self._environment
    self._environment = newEnvironment
    action = S2MLFlattenerAction.NewRestoreEnvironment(previousEnvironment)
    self._PushAction(action)
    self._PushInstantiateActionsForElements(sourceFunction, targetContainer)
    return S2MLFlattenerAction.COMPLETED

# 3.6) Instantiation of elements
# ------------------------------

  def _InstantiateParameter(self, action):
    sourceParameter = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    actualTargetContainer = self._GetActualTargetContainer(sourceParameter, targetContainer)
    if actualTargetContainer==None:
      return S2MLFlattenerAction.FAILED
    sourcePath = sourceParameter.GetPath()
    targetIdentifier = sourcePath.GetLastIdentifier()
    targetParameter = actualTargetContainer.LookForIdentifiedElement(targetIdentifier)
    if targetParameter==None:
      sourceDomain = sourceParameter.GetDomain()
      targetDomain = self._CloneDomain(sourceDomain)
      targetPath = S2MLPath.NewEmptyPath()
      targetPath.AppendIdentifier(targetIdentifier)
      targetParameter = self._targetModel.NewParameter(targetDomain, targetPath, actualTargetContainer)
      targetParameter.SetIdentifierAndInsertInContainerDictionary(targetIdentifier)
    elif targetParameter.GetType()==S2MLType.PARAMETER:
      sourceDomain = sourceParameter.GetDomain()
      targetDomain = self._CloneDomain(sourceDomain)
      targetParameter.SetDomain(targetDomain)
    else:
      absoluteSourcePath = sourceParameter.GetAbsolutePath()
      self.RaiseError("Type mismatch for parameter %s\n" % absoluteSourcePath.ToString())
      return S2MLFlattenerAction.FAILED
    sourceExpression = sourceParameter.GetDefinition()
    targetExpression = self._ContextualizedCopyOfStochasticExpression(sourceExpression, targetContainer, actualTargetContainer)
    if targetExpression==None:
      return S2MLFlattenerAction.FAILED
    targetParameter.SetDefinition(targetExpression)
    return S2MLFlattenerAction.COMPLETED

  def _GetActualTargetContainer(self, sourceElement, targetContainer):
    sourceElementPath = sourceElement.GetPath()
    actualTargetContainerPath = S2MLPath.Copy(sourceElementPath)
    actualTargetContainerPath.RemoveLastIdentifier()
    actualTargetContainer = targetContainer.AccessNamedElement(actualTargetContainerPath)
    if actualTargetContainer==None or actualTargetContainer.IsAtomic():
      absoluteSourcePath = sourceElement.GetAbsolutePath()
      message = "Unable to access parent block of {0:s}\n".format(absoluteSourcePath.ToString())
      self.RaiseError(message)
      return None
    return actualTargetContainer

  def _ContextualizedCopyOfStochasticExpression(self, expression, targetContainer, actualTargetContainer):
    if expression.GetType()==S2MLType.REFERENCE_EXPRESSION:
      sourceReferencePath = expression.GetReferencePath()
      if sourceReferencePath.IsIdentifier():
        identifier = sourceReferencePath.ToIdentifier()
        argument = self.GetValueOfIdentifierAsExpression(identifier)
        if argument!=None:
          environment = self._environment
          self._environment = None
          targetExpression = self._ContextualizedCopyOfStochasticExpression(argument, targetContainer, actualTargetContainer)
          self._environment = environment
          return targetExpression
      targetElement = targetContainer.AccessNamedElement(sourceReferencePath)
      if targetElement==None: # should not happen
        targetReferencePath = S2MLPath.Copy(sourceReferencePath)
      elif sourceReferencePath.IsAbsolute():
        targetReferencePath = targetElement.GetContextualizedAbsolutePath()
      else:
        targetReferencePath = actualTargetContainer.GetRelativePathToOtherElement(targetElement)
      return S2MLStochasticExpression.NewReference(targetReferencePath)
    if expression.GetType()==S2MLType.CONSTANT_EXPRESSION:
      sourceValue = expression.GetValue()
      targetValue = self._targetModel.NewConstant(sourceValue)
      return S2MLStochasticExpression.NewConstant(targetValue)
    if expression.GetType()==S2MLType.WITHIN_EXPRESSION:
      sourceChild = expression.GetChild(0)
      targetChild = self._ContextualizedCopyOfStochasticExpression(sourceChild, targetContainer, actualTargetContainer)
      sourceDomain = expression.GetDomain()
      targetDomain = self._CloneDomain(sourceDomain)
      return S2MLStochasticExpression.NewWithin(targetChild, targetDomain)
    if expression.GetType()==S2MLType.OPERATOR_CALL_EXPRESSION:
      return self._ContextualizedCopyOfOperatorCall(expression, targetContainer, actualTargetContainer)
    childrenCopy = self._ContextualizedCopyOfChildrenOfStochasticExpression(expression, targetContainer, actualTargetContainer)
    if childrenCopy==None:
      return None
    expressionCopy = S2MLStochasticExpression.NewOperation(expression.GetType(), childrenCopy)
    if expression.IsEmpiricalDistribution():
      distribution = expression.GetDistribution()
      distributionCopy = distribution.Copy()
      expressionCopy.SetDistribution(distributionCopy)
    return expressionCopy

  def _ContextualizedCopyOfChildrenOfStochasticExpression(self, expression, targetContainer, actualTargetContainer):
    childrenCopy = []
    for child in expression.GetChildren():
      childCopy = self._ContextualizedCopyOfStochasticExpression(child, targetContainer, actualTargetContainer)
      if childCopy==None:
        return None
      childrenCopy.append(childCopy)
    return childrenCopy

  def _ContextualizedCopyOfOperatorCall(self, expression, targetContainer, actualTargetContainer):
    sourceOperator = self._sourceModel.LookForOperator(expression.GetFunctor())
    if sourceOperator==None:
      absoluteSourcePath = actualParentBlock.GetAbsolutePath()
      message = "In {0:s}, attempt to call undefined operator {1:s}\n".format(absoluteSourcePath.ToString(), expression.GetFunctor())
      self.RaiseError(message)
      return None
    if sourceOperator.GetNumberOfArguments()!=expression.GetNumberOfChildren():
      absoluteSourcePath = actualParentBlock.GetAbsolutePath()
      message = "In {0:s}, wrong number of arguments in application of operator {1:s}\n".format(absoluteSourcePath.ToString(), sourceOperator.GetIdentifier())
      self.RaiseError(message)
      return None
    newEnvironment = S2MLFlattenerEnvironment()
    for index in range(0, sourceOperator.GetNumberOfArguments()):
      formalArgument = sourceOperator.GetArgument(index)
      actualArgument = expression.GetChild(index)
      targetArgument = self._ContextualizedCopyOfStochasticExpression(actualArgument, targetContainer, actualTargetContainer)
      if targetArgument==None:
        return None
      newEnvironment.SetValueOfIdentifier(formalArgument, targetArgument)
    definition = sourceOperator.GetDefinition()
    previousEnvironment = self._environment
    self._environment = newEnvironment
    expression = self._ContextualizedCopyOfStochasticExpression(definition, targetContainer, actualTargetContainer)
    self._environment = previousEnvironment
    return expression

# 3.7) Flattening of models
# -------------------------

  def FlattenElement(self, action):
    sourceElement = action.GetDatum(1)
    if sourceElement.GetType()==S2MLType.MODEL:
      status = self._FlattenModel(action)
    elif sourceElement.GetType()==S2MLType.DOMAIN_DECLARATION:
      status = self._FlattenDomainDeclaration(action)
    elif sourceElement.GetType()==S2MLType.BLOCK:
      status = self._FlattenBlock(action)
    elif sourceElement.GetType()==S2MLType.CLASS:
      status = S2MLFlattenerAction.COMPLETED
    elif sourceElement.GetType()==S2MLType.INSTANCE:
      status = S2MLFlattenerAction.COMPLETED
    elif sourceElement.GetType()==S2MLType.CLONES:
      status = S2MLFlattenerAction.COMPLETED
    elif sourceElement.GetType()==S2MLType.EXTENDS:
      status = S2MLFlattenerAction.COMPLETED
    elif sourceElement.GetType()==S2MLType.PARAMETER:
      status = self._FlattenParameter(action)
    else:
      status = S2MLFlattenerAction.FAILED
    return status

  def _FlattenModel(self, action):
    sourceModel = action.GetDatum(0)
    targetContainer = action.GetDatum(2)
    self._PushFlattenActionsForElements(sourceModel, targetContainer)
    return S2MLFlattenerAction.COMPLETED

  def _PushFlattenActionsForElements(self, sourceContainer, targetContainer):
    for sourceElement in reversed(sourceContainer.GetElements()):
      self._PushFlattenActionForElement(sourceContainer, sourceElement, targetContainer)

  def _PushFlattenActionForElement(self, sourceContainer, sourceElement, targetContainer):
    action = S2MLFlattenerAction.NewFlattenElement(sourceContainer, sourceElement,  targetContainer)
    self._PushAction(action)

  def _FlattenDomainDeclaration(self, action):
    sourceContainer = action.GetDatum(0)
    sourceDeclaration = action.GetDatum(1)
    targetContainer = action.GetDatum(2)
    sourceDomain = sourceDeclaration.GetDomain()
    targetDomain = self._CloneDomain(sourceDomain)
    targetDomain = self._targetModel.NewDomainDeclaration(sourceDeclaration.GetIdentifier(), targetDomain, targetContainer)
    return S2MLFlattenerAction.COMPLETED

  def _FlattenBlock(self, action):
    sourceContainer = action.GetDatum(0)
    sourceBlock = action.GetDatum(1)
    targetContainer = action.GetDatum(2)
    if sourceContainer==self._sourceModel:
      sourcePath = sourceBlock.GetPath()
      targetPath = S2MLPath.Copy(sourcePath)
      targetBlock = self._targetModel.NewBlock(targetPath, targetContainer)
      self._PushFlattenActionsForElements(sourceBlock, targetBlock)
    else:
      self._PushFlattenActionsForElements(sourceBlock, targetContainer)
    return S2MLFlattenerAction.COMPLETED

  def _FlattenParameter(self, action):
    sourceContainer = action.GetDatum(0)
    sourceParameter = action.GetDatum(1)
    targetContainer = action.GetDatum(2)
    sourcePath = sourceParameter.GetPath()
    actualSourceParameter = sourceContainer.AccessNamedElement(sourcePath)
    if actualSourceParameter==None:
      self.RaiseError("This should not happen")
      return assisFlattenerAction.FAILED
    targetPath = actualSourceParameter.GetAbsolutePath()
    targetPath.RemoveFirstIdentifier()
    targetIdentifier = targetPath.ToString()
    targetParameter = targetContainer.LookForIdentifiedElement(targetIdentifier)
    if targetParameter==None:
      sourceDomain = sourceParameter.GetDomain()
      targetDomain = self._CloneDomain(sourceDomain)
      targetParameter = self._targetModel.NewParameter(targetDomain, targetPath, targetContainer)
      targetParameter.SetIdentifierAndInsertInContainerDictionary(targetIdentifier)
    sourceExpression = sourceParameter.GetDefinition()
    targetExpression = self._FlattenStochasticExpression(sourceContainer, sourceExpression)
    if targetExpression==None:
      return S2MLFlattenerAction.FAILED
    targetParameter.SetDefinition(targetExpression)
    return S2MLFlattenerAction.COMPLETED

  def _FlattenStochasticExpression(self, sourceContainer, sourceExpression):
    if sourceExpression.GetType()==S2MLType.REFERENCE_EXPRESSION:
      sourceReferencePath = sourceExpression.GetReferencePath()
      sourceReferredElement = sourceContainer.AccessNamedElement(sourceReferencePath)
      #if sourceReferredElement==None or not S2MLModel.IsParameter(sourceReferredElement):
      if sourceReferredElement==None:
        absoluteSourcePath = sourceContainer.GetAbsolutePath()
        message = "In {0:s}, undefined reference {1:s}\n".format(absoluteSourcePath.ToString(), sourceReferencePath.ToString())
        self.RaiseError(message)
        return None
      targetReferencePath = sourceReferredElement.GetAbsolutePath()
      targetReferencePath.RemoveFirstIdentifier()
      targetReferencePath.MakeIdentifier()
      return S2MLStochasticExpression.NewReference(targetReferencePath)
    if sourceExpression.GetType()==S2MLType.CONSTANT_EXPRESSION:
      value = sourceExpression.GetValue()
      return S2MLStochasticExpression.NewConstant(value)
    if sourceExpression.GetType()==S2MLType.WITHIN_EXPRESSION:
      sourceChild = sourceExpression.GetChild(0)
      targetChild = self._FlattenStochasticExpression(sourceContainer, sourceChild)
      sourceDomain = sourceExpression.GetDomain()
      targetDomain = self._CloneDomain(sourceDomain)
      return S2MLStochasticExpression.NewWithin(targetChild, targetDomain)
    targetChildren = self._FlattenChildrenOfStochasticExpression(sourceContainer, sourceExpression)
    if targetChildren==None:
      return None
    targetExpression = S2MLStochasticExpression.NewOperation(sourceExpression.GetType(), targetChildren)
    if sourceExpression.IsEmpiricalDistribution():
      sourceDistribution = sourceExpression.GetDistribution()
      targetDistribution = sourceDistribution.Copy()
      targetExpression.SetDistribution(targetDistribution)
    return targetExpression

  def _FlattenChildrenOfStochasticExpression(self, sourceContainer, sourceExpression):
    targetChildren = []
    for sourceChild in sourceExpression.GetChildren():
      targetChild = self._FlattenStochasticExpression(sourceContainer, sourceChild)
      if targetChild==None:
        return None
      targetChildren.append(targetChild)
    return targetChildren
