# Janos version 1.2.1
# JanosModelChecker.py
# by Antoine Rauzy
# Copyrights (c) 2018-2022 AltaRica Association
#
# This file is part of Janos, a pedagogical stochastic simulator.
# Janos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Janos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Janos.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Undefined references
# 4) Circular definitions
# 5) Types
# 6) Tool specific checks

# 1) Imported modules
# -------------------

from S2MLCore import *

from JanosModels import *

# 2) Creator
# ----------

class JanosModelChecker(S2MLModelChecker):
  def __init__(self, engine):
    S2MLModelChecker.__init__(self, engine)

# 3) Undefined references
# -----------------------

  def CheckReferencesInBlock(self, block):
    for element in block.GetElements():
      if element.GetType()==JanosType.BLOCK:
        previousCheckedBlock = self._checkedBlock
        self._checkedBlock = element
        self.CheckReferencesInBlock(element)
        self._checkedBlock = previousCheckedBlock
      elif element.GetType()==JanosType.VARIABLE:
        self._checkedElement = element
        self._CheckReferencesInAttributesOfElement(element)
        self._checkedElement = None
      elif element.GetType()==JanosType.ASSIGNMENT:
        self._CheckReferencesInEquation(element)
      elif element.GetType()==JanosType.DOUBLE_ASSIGNMENT:
        self._CheckReferencesInEquation(element)
      elif element.GetType()==JanosType.PARAMETER:
        self._checkedElement = element
        self._CheckReferencesInStochasticExpression(element.GetDefinition())
        self._checkedElement = None
      elif element.GetType()==JanosType.OBSERVER:
        self._checkedElement = element
        self._CheckReferencesInStochasticExpression(element.GetDefinition())
        self._checkedElement = None
    return self._errors

  def _CheckReferencesInEquation(self, equation):
    leftMember = equation.GetMember(0)
    referencePath = leftMember.GetReferencePath()
    element = self._checkedBlock.AccessNamedElement(referencePath)
    if element==None:
      referenceIdentifier = referencePath.ToString()
      element = self._checkedBlock.LookForIdentifiedElement(referenceIdentifier)
    if element==None:
      self.RaiseError("In block {0:s}, undefined reference as left member of an assignment {1:s}".format(self._checkedBlock.GetIdentifier(), referencePath.ToString()))
      self._errors += 1
      return
    self._checkedElement = element
    rightMember = equation.GetMember(1)
    self._CheckReferencesInStochasticExpression(rightMember)
    self._checkedElement = None

# 4) Circular definitions
# -----------------------
# Circular definition are checked by the rewriter

# 5) Types
# --------
# Not implemented yet

# 6) Tool specific checks
# -----------------------

  def ToolSpecificCheckModel(self, model):
    # self.CheckInitialValuesOfVariables(model)
    pass

  def CheckInitialValuesOfVariables(self, block):
    for element in block.GetElements():
      if element.GetType()==JanosType.BLOCK:
        previousCheckedBlock = self._checkedBlock
        self._checkedBlock = element
        self.CheckInitialValuesOfVariables(element)
        self._checkedBlock = previousCheckedBlock
      elif element.GetType()==JanosType.VARIABLE:
        initialValue = element.GetAttribute("init")
        if initialValue==None:
          self.RaiseError("Missing initial value (init attribute) for variable {0:s}".format(element.GetIdentifier()))
          self._errors += 1
    return self._errors


